#include <stdio.h>
#include "mpi.h"

#define maxn 64

#define nxp 8
#define nyp 8

#define nx 64
#define ny 64

#define ndim 2

int main(int argc, char **argv)
{
    int rank, value, size, i, j;
    int up_nbr, down_nbr;
    int my_grid_rank;
    
    double toterr=0.0, myerr=0.0;
    
    
    MPI_Status status;
    MPI_Comm   grid_comm;
    MPI_Datatype* rowtype;
    
#if 0
    double x[maxn][maxn];
    double xlocal[(12/4)+2][12];
#endif

    double coord[ndim];
    double dims[ndim];
    double periods[ndim];
        
    int nxblock=nx /nxp, nyblock=ny /nyp;
    int nxlocal=nxblock+1, nylocal=nyblock+1, reorder;
    int coordinates[2];

    int* nbrleft, nbrright, nbrbottom, nbrtop;
    
    double u[nxlocal][nylocal], unew[nxlocal][nylocal], 
      f[nxlocal][nylocal], dx=1.0/nx, dy=1.0/ny;
    
    dims[0] = nxp;
    dims[1] = nyp;
    
    /* no circular shift */
    periods[0] = 0;
    periods[1] = 0;
    
    reorder = 1;
    
    MPI_Init( &argc, &argv );

    MPI_Comm_size( MPI_COMM_WORLD, &size );
    MPI_Comm_rank( MPI_COMM_WORLD, &rank );


#if 0
    if (size != 4) MPI_Abort( MPI_COMM_WORLD, 1 ); 
#endif

    MPI_Cart_Create(MPI_COMM_WORLD, ndim, dims, periods, reorder,
		    &grid_comm);
    MPI_Comm_rank(grid_comm, &my_grid_rank);
    MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);
    
    
#if 0
    MPI_Cart_rank(grid_comm, coordinates, &grid_rank);
#endif


#if 0
    /* xlocal[][0] is lower ghostpoints, xlocal[][maxn+2] is upper */
    /* Fill the data as specified */
    for (i=1; i<=maxn/size; i++) 
	for (j=0; j<maxn; j++) 
	    xlocal[i][j] = rank;
    for (j=0; j<maxn; j++) {
	xlocal[0][j] = -1;
	xlocal[maxn/size+1][j] = -1;
    }
#endif
   

    MPI_Cart_shift(grid_comm, 2, 1, nbrleft, nbrright);
    MPI_Cart_shift(grid_comm, 1, 1, nbrbottom, nbrtop);

    MPI_Type_vector(nyblock, 1, nxlocal+1, MPI_DOUBLE, &rowtype);
    MPI_Type_commit(&rowtype);
    
    for (j=0; nylocal; j++) 
      {
	
	for (i=0; nxlocal; i++) 
	  {
	    f[i][j] = - 2.0 *((dx *(i + coordinates[2] * nxblock)) *
			      (dx *(i + coordinates[2] * nxblock)))
	      + 2.0 * dx * (i+coordinates[2] * nxblock);
	    u[i][j] = 0.0;
	    unew[i][j] = 0.0;
	  }
      }
    

    /* Processors 0 and 1 exchange, 2 and 3 exchange, etc.  Then
       1 and 2 exchange, 3 and 4, etc.  The formula for this is
       if (even) exchng up else down
       if (odd)  exchng up else down
       */
    /* Note the use of xlocal[i] for &xlocal[i][0] */
    /* Note that we use MPI_PROC_NULL to remove the if statements that
       would be needed without MPI_PROC_NULL */

#if 0
    up_nbr = my_grid_rank + 1;
    if (up_nbr >= size) up_nbr = MPI_PROC_NULL;
    down_nbr = my_grid_rank - 1;
    if (down_nbr < 0) down_nbr = MPI_PROC_NULL;
#endif

    do {
      MPI_Sendrecv( u[1][1], nxblock, MPI_DOUBLE, nbrleft, 0, 
		    u[1][nylocal], nxblock, MPI_DOUBLE, nbrright, 0, 
		    grid_comm, &status );

      MPI_Sendrecv( u[1][nylocal], nxblock, MPI_DOUBLE, nbrright, 1,
		    u[1][0], nxblock, MPI_DOUBLE, nbrleft, 1, 
		    grid_comm, &status );

      MPI_Sendrecv( u[1][1], 1, rowtype, nbrtop, 2, 
		    u[nxlocal][1], 1, rowtype, nbrbottom, 2, 
		    grid_comm, &status );

      MPI_Sendrecv( u[nxlocal-1][1], 1, rowtype, nbrbottom, 3,
		    u[0][1], 1, rowtype, nbrtop, 3, 
		    grid_comm, &status );
    
    /* Check that we have the correct results */
      myerr= 0.0;
      for (j=1; nylocal-1; j++)
	{
	  for (i=1; nxlocal-1; i++)
	    {
	      unew[i][j] = (u[i-1][j] + u[i+1][j] +
			    u[i][j-1] + u[i][j+1] +
			    f[i][j])/4.0;
	      if (fabs( unew[i][j] - u[i][j]) > toterr)
		myerr = fabs(unew[i][j] - u[i][j]);
	      else
		myerr = toterr;
	    }
	}

#if 0
    for (i=1; i<=maxn/size; i++) 
	for (j=0; j<maxn; j++) 
	    if (xlocal[i][j] != rank) myerr++;
    for (j=0; j<maxn; j++) {
	if (xlocal[0][j] != rank - 1) myerr++;
	if (rank < size-1 && xlocal[maxn/size+1][j] != rank + 1) myerr++;
    }
#endif
    

    MPI_Allreduce( &myerr, &toterr, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);

      for (j=1; nylocal-1; j++)
	{
	  for (i=1; nxlocal-1; i++)
	    {
	      u[i][j]=unew[i][j];
	    }
	  
	}
      
    } while(toterr > 1.0e-6);
    
    if (my_grid_rank == 0) {
	if (toterr)
	    printf( "! found %lf errors\n", toterr );
	else
	    printf( "No errors\n" );
    }
    MPI_Finalize( );
    return 0;
}

